[
  {'base_name': 'one_service',
   'service_type': 'cone',
   'access_url': 'http://vo.cool_archive_cone.org/vo/conesearch',
   },
  {'base_name': 'another_service',
   'service_type': 'cone',
   'access_url': 'http://www.another_archive.org/big_catalog/conesearch',
   }
]



[
    {'base_name': 'HEASARC_swiftmastr',
     'service_type': 'tap',
     'access_url': 'https://heasarc.gsfc.nasa.gov/xamin/vo/tap',
     'adql':'''
SELECT
   *
   FROM swiftmastr
   WHERE 
      1=CONTAINS(POINT('ICRS', ra, dec),
                 CIRCLE('ICRS', {}, {}, {}))
    '''
     }
]



from servicemon.plugin_support import AbstractTimedQuery


class UserQuery1(AbstractTimedQuery, plugin_name='userquery1',
                 description="A query plug-in in the user's dir that does nothing."):
    pass



import requests

from servicemon.plugin_support import AbstractResultWriter


class ImportTester(AbstractResultWriter, plugin_name='import_tester',
                   description='A writer in the default plug-in dir that does nothing but try an import.'):

    def begin(self, args, **kwargs):

        # Test that the import of requests worked.
        print(f'requests = {requests}')

    def one_result(self, stats):
        pass

    def end(self):
        pass



import re
from importlib import import_module
from abc import ABC, abstractmethod

from pathlib import Path

from servicemon import builtin_plugins


class SmPluginSupport(ABC):

    _subclasses = {}
    _default_user_plugin_dir = 'plugins'

    @classmethod
    def __init_subclass__(cls, plugin_name=None, description='', **kwargs):
        super().__init_subclass__(**kwargs)
        if plugin_name is None:
            plugin_name = cls.__name__
        cls._subclasses[plugin_name] = SmPluginDesc(plugin_name, cls, description)
        # {'cls': cls, 'description': description}

    @classmethod
    def load_builtin_plugins(cls):
        for dirstr in builtin_plugins.__path__:
            path = Path(dirstr)
            for child in path.iterdir():
                if cls.__is_python_file(child):
                    module = 'servicemon.builtin_plugins.' + child.stem
                    import_module(module)

    @classmethod
    def load_plugins(cls, plugins=None):
        if plugins is None:
            plugins = cls._default_user_plugin_dir

        plugin_path = Path(plugins)
        if cls.__is_python_file(plugin_path):
            cls.__load_file(plugin_path)
        elif plugin_path.is_dir():
            for child in plugin_path.iterdir():
                if cls.__is_python_file(child):
                    cls.__load_file(child)
        else:
            # Only error on non-default file/directory.
            if plugins != cls._default_user_plugin_dir:
                raise FileNotFoundError(f"Plugin file/dir doesn't exist: {plugin_path}")

    @classmethod
    def list_plugins(cls):
        for k in cls._subclasses:
            print(f'subclass key = {k}, value = {cls._subclasses[k]}')

    @classmethod
    def get_plugin(cls, plugin_name):
        plugin = cls._subclasses.get(plugin_name, None)
        return plugin

    @classmethod
    def get_plugin_from_spec(cls, spec):
        parsed = cls.parse_spec(spec)
        plugin = cls.get_plugin(parsed['plugin_name'])
        if plugin is None:
            raise ValueError(f'No plugin found for spec: {spec}')
        full_plugin = SmPluginDesc(plugin.name, plugin.cls, plugin.description, **parsed['kwargs'])
        return full_plugin

    @classmethod
    def parse_spec(cls, spec):
        """
        The plugin specification is a string of the form:

        <plugin> :== <plugin_name> | <plugin_name> ":" <kwargs>
        <kwargs> :== <kwarg> | <kwarg> "," <kwargs>
        <kwarg> :== <key> "=" <value>
        <plugin_name> :== string with no ":"
        <key> :== string with no "=" or ","
        <value> :== string with no ","

        Whitespace will be ignored.

        """
        parsed = None
        stripped = re.sub(r"\s+", "", spec, flags=re.UNICODE)

        # Get the plug-in name.
        matches = re.match(r"^(?P<plugin_name>[^:]*)", stripped)
        if matches is not None:
            next_idx = 0
            parsed = {"plugin_name": matches.group('plugin_name'),
                      "kwargs": {}}

            # Get the keyword arguments.
            while matches is not None:
                next_idx += matches.end() + 1   # Skip the colon or comma, if any.
                matches = re.match(r"(?P<key>[^=,]+)=(?P<value>[^,]+)",
                                   stripped[next_idx:])
                if matches is not None:
                    parsed['kwargs'][matches.group('key')] = matches.group('value')

            # If we didn't finish matching the whole spec, the spec is not valid.
            if next_idx - 1 != len(stripped):
                raise ValueError(f'Invalid plug-in specification: "{spec}"')

        return parsed

    @classmethod
    def __load_file(cls, path: Path):
        python_code = path.read_text(encoding='utf-8')
        compiled_code = compile(python_code, path.name, 'exec')
        exec(compiled_code, globals())

    @classmethod
    def __is_python_file(cls, path: Path):
        is_python_file = path.is_file() and path.suffix == '.py'
        return is_python_file


class SmPluginDesc():
    def __init__(self, name, cls, description, **kwargs):
        self.name = name
        self.cls = cls
        self.description = description
        self.kwargs = kwargs


class AbstractResultWriter(SmPluginSupport):

    _subclasses = {}

    @abstractmethod
    def begin(self, args, **kwargs):
        """
        **args** : argparse.Namespace
            the result of an argparse.ArgumentParser's parse_args().

        **kwargs** : dict
            keyword args from the plug-in specification.
        """
        pass

    @abstractmethod
    def one_result(self, stats):
        """
        **stats** : obj
            an object with the following methods:

            **columns()** : list of str
                list of output column names

            **row_values()** : dict
                dict of output values, one key per column name
        """
        pass

    @abstractmethod
    def end(self):
        pass


class AbstractTimedQuery(SmPluginSupport):

    _subclasses = {}



import os
import pathlib
import warnings
import traceback
import logging
import html
import requests
import sys

from codetiming import Timer
import servicemon
from astropy.coordinates import SkyCoord
from astropy.table import Table
from astroquery.utils import parse_coordinates

from .query_stats import Interval, QueryStats
from .pyvo_wrappers import TAPServiceSM


def time_this(interval_name):
    def time_this_decorator(func):
        def wrapper(*args, **kwargs):
            interval = Interval(interval_name)
            result = func(*args, **kwargs)
            interval.close()
            args[0].stats.add_interval(interval)

            return result
        return wrapper
    return time_this_decorator


def compute_user_agent(specified_agent):
    user_agent = specified_agent
    if user_agent is None:
        vi = sys.version_info
        product = f'servicemon/{servicemon.__version__}'
        comment = '(IVOA-monitor https://github.com/NASA-NAVO/servicemon)'
        python_version = f'Python/{vi.major}.{vi.minor}.{vi.major}'
        user_agent = f'{product} {comment} {python_version}'
    return user_agent


class Query():
    """
    """

    def __init__(self, service, coords, radius, out_dir, use_subdir=True,
                 agent=None, tap_mode='async', save_results=True,
                 verbose=False):
        self._save_results = save_results

        self._timer = Timer('query_total', logger=None)
        self._timer.timers.clear()

        self.__agent = compute_user_agent(agent)
        self._tap_mode = tap_mode
        self._service = service
        self._base_name = self._compute_base_name()
        self._service_type = self._compute_service_type()
        self._use_subdir = use_subdir

        self._orig_coords = coords
        self._orig_radius = radius
        self._coords = self._compute_coords()
        self._adql = self._compute_adql()
        self._access_url = self._compute_access_url()

        # Add cone params to xcone access_url
        if self._service_type == 'xcone':
            ra, dec, radius = self._get_ra_dec_radius()
            self._access_url = self._access_url.format(ra, dec, radius)

        if self._use_subdir:
            self._out_path = pathlib.Path(f'{out_dir}/{self._base_name}')
        else:
            self._out_path = pathlib.Path(out_dir)
        self._verbose = verbose

        self._query_params = self._compute_query_params()
        self._query_name = self._compute_query_name()
        self._filename = self._out_path / (self._query_name + '.xml')

        self._stats = QueryStats(
            self._query_name, self._base_name, self._service_type,
            self._access_url, self._query_params, self._result_meta_attrs())

    @property
    def stats(self):
        return self._stats

    def run(self):
        with self._timer:
            if self._service_type == 'cone':
                response = self.do_cone_query()
                self.stream_to_file(response)
            if self._service_type == 'xcone':
                response = self.do_xcone_query()
                self.stream_to_file(response)
            elif self._service_type == 'tap':
                tap_service = TAPServiceSM(self._access_url)
                if self._tap_mode == 'async':
                    response = self.do_tap_query_async_pyvo(tap_service)
                else:
                    response = self.do_tap_query_pyvo(tap_service)
                self.stream_to_file(response)

        self.gather_response_metadata(response)

    @time_this('do_query')
    def do_tap_query_async_pyvo(self, tap_service):
        response = tap_service.run_async_timed(self._adql, streamable_response=True)
        return response

    @time_this('do_query')
    def do_tap_query_pyvo(self, tap_service):
        response = tap_service.run_sync_timed(self._adql, streamable_response=True)
        return response

    @time_this('do_query')
    def do_cone_query(self):
        response = self.do_request(self._access_url, self._query_params)
        return response

    @time_this('do_query')
    def do_xcone_query(self):
        response = self.do_request(self._access_url)
        return response

    @time_this('stream_to_file')
    def stream_to_file(self, response):
        os.makedirs(os.path.dirname(self._filename), exist_ok=True)
        with open(self._filename, 'wb+') as fd:
            for chunk in response.iter_content(chunk_size=8096):
                fd.write(chunk)

    def compute_headers(self):
        headers = requests.utils.default_headers()
        if self.__agent is not None:
            headers.update({
                'User-Agent': self.__agent
            })
        return headers

    def do_request(self, url, params=None, agent=None):
        headers = self.compute_headers()

        response = requests.get(url, params, headers=headers, stream=True)
        return response

    def _result_meta_attrs(self):
        return ['status', 'size', 'num_rows', 'num_columns']

    def _handle_exc(self, msg, trace=False):
        self._stats.errmsg = self._stats.errmsg + msg
        if trace:
            traceback.print_exc()
        else:
            logging.error(msg)

    def gather_response_metadata(self, response):
        """
        response:  Either an http.client.HTTPResponse or a yyy
        """
        # Add timings to stats intervals.
        timers = self._timer.timers
        for items in timers.items():
            self._stats.add_interval(Interval(items[0], items[1]))

        result_meta = dict.fromkeys(self._result_meta_attrs())
        result_meta['status'] = response.status_code

        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                t = Table.read(self._filename, format='votable')

            result_meta['size'] = os.path.getsize(self._filename)
            result_meta['num_rows'] = len(t)
            result_meta['num_columns'] = len(t.columns)

            if not self._save_results:
                os.remove(self._filename)
        except Exception as e:
            msg = f'In {self._query_name}, error reading result table: {repr(e)}'
            self._handle_exc(msg)
        finally:
            self._stats.result_meta = result_meta

    def _compute_base_name(self):
        base_name = self.getval(self._service, 'base_name', 'Unnamed')
        return base_name

    def _compute_service_type(self):
        service_type = self.getval(self._service, 'service_type')
        return service_type

    def _compute_access_url(self):
        access_url = self.getval(self._service, 'access_url')
        if access_url is None:
            raise ValueError('service must have an access_url')
        access_url = html.unescape(access_url)

        return access_url

    def _compute_coords(self):
        # Get the RA and Dec from in_coords.
        in_coords = self._orig_coords
        coords = in_coords
        if in_coords is not None:
            if ((type(in_coords) is tuple or type(in_coords) is list) and
                    len(in_coords) == 2):
                coords = SkyCoord(in_coords[0], in_coords[1], frame="icrs",
                                  unit="deg")
            elif type(in_coords) is str:
                coords = parse_coordinates(in_coords)
            elif type(in_coords) is not SkyCoord:
                raise ValueError(f"Cannot parse input coordinates {in_coords}")

        return coords

    def _compute_adql(self):
        adql = self.getval(self._service, 'ADQL', '')
        if adql == '':
            adql = self.getval(self._service, 'adql', '')

        # coords will be None on a replay,
        # since this substitution has already happened.
        if self._coords is not None:
            adql = adql.format(self._coords.ra.deg, self._coords.dec.deg,
                               self._orig_radius)

        return adql

    def _get_ra_dec_radius(self):
        if self._coords is not None:
            ra = self._coords.ra.deg
            dec = self._coords.dec.deg
            radius = self._orig_radius
        else:
            ra = self.getval(self._service, 'RA', None)
            dec = self.getval(self._service, 'DEC', None)
            radius = self.getval(self._service, 'SR', None)

        # Hack because Chandra TAP service does arminutes instead of degrees.
        if self._access_url.startswith('http://cda.harvard.edu/csctap'):
            radius = 60 * radius

        return ra, dec, radius

    def _compute_query_params(self):
        ra, dec, radius = self._get_ra_dec_radius()
        params = {
            'RA': ra,
            'DEC': dec,
            'SR': radius
        }
        if self._service_type == 'tap':
            params['ADQL'] = self.fix_white(self._adql)

        return params

    def _compute_query_name(self):
        name = f'{self._base_name}_{self._service_type}'
        if self._service_type == 'tap':
            name += f'-{self._tap_mode}'
        ra = self._query_params.get('RA')
        dec = self._query_params.get('DEC')
        sr = self._query_params.get('SR')
        if ra is not None and dec is not None and sr is not None:
            name += f'_{ra}_{dec}_{sr}'

        return name

    def getval(self, obj, key, default=None):
        """
        Gets the value as either an attribute val or key val.
        """
        val = getattr(obj, key, None)
        if val is None:
            try:
                val = obj[key]
            except KeyError:
                val = default
        return val

    def fix_white(self, s):
        fixed = " ".join(s.split())
        return fixed



from servicemon.plugin_support import AbstractResultWriter


class CannotInstantiate(AbstractResultWriter, plugin_name='cannot_make_reader_instance',
                        description="A writer in the user's plug-in dir that cannot be instatiated"
                        " because it doesn't implement the end() method from the abstract base class."):

    def begin(self, args, **kwargs):
        pass

    def one_result(self, stats):
        pass



import csv
import os
import sys
from pathlib import Path
from datetime import datetime

from servicemon.plugin_support import AbstractResultWriter


class CsvResultWriter(AbstractResultWriter, plugin_name='csv_writer',
                      description='Writes results to a csv file.'):

    def begin(self, args, outfile=None):
        self._first_stat = True
        self._outfile_path = self._compute_outfile_path(args, outfile=outfile)

        # Create the output dir if needed.
        if self._outfile_path is not None:
            os.makedirs(self._outfile_path.parent, exist_ok=True)

    def end(self):
        pass

    def one_result(self, stats):

        if self._outfile_path is not None:
            with open(self._outfile_path, 'a+') as file:
                self._output_stats_row_to_file(stats, file)
        else:
            self._output_stats_row_to_file(stats, sys.stdout)

    def _output_stats_row_to_file(self, stats, file):
        writer = csv.DictWriter(file, dialect='excel',
                                fieldnames=stats.columns())
        if self._first_stat:
            self._first_stat = False
            writer.writeheader()
        writer.writerow(stats.row_values())

    def _compute_outfile_path(self, args, outfile=None):
        """
        If outfile is not None, use it.  Otherwise compute the output file
        name from name of the services file supplied in the args provided.
        """
        result_path = None
        if outfile is not None:
            # Leaving result_path None will cause output to go to stdout.
            if outfile != 'stdout':
                result_path = Path(outfile)
        else:
            now = datetime.now()
            dtstr = now.strftime('%Y-%m-%d %H:%M:%S.%f')

            base_services_name = Path(args.services).stem
            result_name = f'{base_services_name}_{dtstr}.csv'

            result_path = Path(args.result_dir) / Path(result_name)

        return result_path



import pytest
from pathlib import Path

from astropy.utils.data import get_pkg_data_filename

from servicemon.plugin_support import SmPluginSupport, AbstractResultWriter, AbstractTimedQuery
from servicemon.query_runner import _parse_query


def load_builtins(capsys):
    SmPluginSupport.load_builtin_plugins()


def load_from_default_user_dir(capsys):
    data_file = get_pkg_data_filename('data/default_user_plugin_dir/basic_writer.py')
    data_dir = Path(data_file).parent
    SmPluginSupport._default_user_plugin_dir = data_dir

    SmPluginSupport.load_plugins()


def load_from_user_dir(capsys):
    plugin_file = get_pkg_data_filename('data/user_plugin_dir/userquery1.py')
    plugin_dir = Path(plugin_file).parent
    SmPluginSupport.load_plugins(plugins=plugin_dir)


def load_from_user_file(capsys):
    plugin_file = get_pkg_data_filename('data/sample_writer.py')
    SmPluginSupport.load_plugins(plugins=plugin_file)


def test_load_plugins(capsys):
    load_builtins(capsys)
    load_from_default_user_dir(capsys)
    load_from_user_dir(capsys)
    load_from_user_file(capsys)

    assert len(AbstractResultWriter._subclasses) == 7  # lengths include the Abstract* base classes.
    assert len(AbstractTimedQuery._subclasses) == 4

    # Check built-ins
    pi = AbstractResultWriter.get_plugin_from_spec('csv_writer:outfile=somedir/somefile.csv')
    assert 'Writes results to a csv file.' in pi.description
    assert pi.cls.__name__ == 'CsvResultWriter'
    assert pi.kwargs['outfile'] == 'somedir/somefile.csv'

    # Check user default dir
    AbstractResultWriter.list_plugins()

    pi = AbstractResultWriter.get_plugin_from_spec('basic_writer')
    assert 'Another writer in the default plug-in dir' in pi.description
    assert pi.cls.__name__ == 'BasicWriter'
    assert pi.kwargs == {}
    _ = pi.cls()   # Instantiates fine since it implements the abstract methods.

    pi = AbstractTimedQuery.get_plugin_from_spec('defquery1:arg1=abc,arg2=def')
    assert 'A query plug-in ' in pi.description
    assert pi.cls.__name__ == 'DefQuery1'
    assert pi.kwargs['arg1'] == 'abc'
    assert pi.kwargs['arg2'] == 'def'

    # Check user plugin dir
    pi = AbstractTimedQuery.get_plugin_from_spec('userquery2')
    assert pi.cls.__name__ == 'UserQuery2'

    pi = AbstractResultWriter.get_plugin_from_spec('cannot_make_reader_instance')
    assert pi.cls.__name__ == 'CannotInstantiate'
    with pytest.raises(TypeError) as record:
        _ = pi.cls()
        assert ("Can't instantiate abstract class CannotInstantiate with abstract methods end"
                in str(record[0].message))

    # Check user file
    pi = AbstractResultWriter.get_plugin_from_spec('sample-writer')
    assert pi.cls.__name__ == 'SampleResultWriter'


def test_imports(capsys):
    load_from_default_user_dir(capsys)
    pi = AbstractResultWriter.get_plugin_from_spec('import_tester')
    pi_instance = pi.cls()

    # Make sure we can run begin() which tries to print out the imported 'requests' module.
    args = _parse_query([
        'fake_services_file.py',
        '--cone_file', 'my_cones.py'
    ])
    pi_instance.begin(args)

    captured = capsys.readouterr()
    output = str(captured.out)
    assert "module 'requests'" in output



[
    {'base_name': 'PanSTARRS',
     'service_type': 'tap',
     'access_url': 'http://vao.stsci.edu/PS1DR2/tapservice.aspx',
     'adql':'''
   SELECT objID, RAMean, DecMean, nDetections, ng, nr, ni, nz, ny, gMeanPSFMag,
   rMeanPSFMag, iMeanPSFMag, zMeanPSFMag, yMeanPSFMag
   FROM dbo.MeanObjectView
   WHERE
   CONTAINS(POINT('ICRS', RAMean, DecMean),CIRCLE('ICRS',{},{},{}))=1
    '''
     }
]



import pytest
from servicemon.cone import Cone, _parse_cone_args, conegen_defaults


def errstr(capsys):
    """
    A shortcut with a side effect:  Resets captured stdout and stderr.
    """
    captured = capsys.readouterr()
    result = str(captured.err)
    return result


def validate_cone(num_points, min_radius, max_radius):
    gen = Cone.generate_random(num_points, min_radius, max_radius)
    prev = {'ra': 0, 'dec': 0, 'radius': 0}
    points = list(gen)
    assert len(points) == num_points
    for cone in gen:
        assert 0 <= cone['ra'] < 360
        assert -90 <= cone['dec'] < 90
        assert min_radius <= cone['radius'] <= max_radius

        assert cone['ra'] != prev['ra']
        assert cone['dec'] != prev['dec']
        assert cone['radius'] != prev['radius']

        prev = cone


def test_random():
    validate_cone(100, 0.01, 1.5)
    validate_cone(10000, 14.5, 102)


def test_errors():
    with pytest.raises(ValueError) as e_info:
        Cone.random_cone(2, 1)
    assert ('min-radius must be in the range [0,max_radius).'
            in str(e_info.value))

    with pytest.raises(ValueError) as e_info:
        Cone.generate_random(20, 3, 1)
    assert ('min-radius must be in the range [0,max_radius).'
            in str(e_info.value))

    with pytest.raises(ValueError) as e_info:
        Cone.generate_random(-14, 6, 7)
    assert ('num_points must be a positive number.'
            in str(e_info.value))


def test_parse_args(capsys):

    # With defaults
    args = _parse_cone_args([
        'output_cone_file.py',
        '--num_cones', '27'
    ])
    assert args.outfile == 'output_cone_file.py'
    assert args.num_cones == 27
    assert args.min_radius == conegen_defaults['min_radius']
    assert args.max_radius == conegen_defaults['max_radius']

    # Without defaults
    args = _parse_cone_args([
        'output_cone_file.py',
        '--num_cones', '438',
        '--min_radius', '12.3',
        '--max_radius', '45.6'
    ])
    assert args.outfile == 'output_cone_file.py'
    assert args.num_cones == 438
    assert args.min_radius == 12.3
    assert args.max_radius == 45.6

    # Need outfile
    with pytest.raises(SystemExit):
        _ = _parse_cone_args([
            '--num_cones', '438',
            '--min_radius', '12.3',
            '--max_radius', '45.6'
        ])
    assert ('the following arguments are required: outfile'
            in errstr(capsys))

    # Need num_cones
    with pytest.raises(SystemExit):
        _ = _parse_cone_args([
            'some_file.py',
            '--min_radius', '12.3',
            '--max_radius', '45.6'
        ])
    assert ('the following arguments are required: --num_cones'
            in errstr(capsys))

    # Bad types
    with pytest.raises(SystemExit):
        _ = _parse_cone_args([
            'some_file.py',
            '--num_cones', 'notanint'
        ])
    assert ('--num_cones: invalid int value'
            in errstr(capsys))
    with pytest.raises(SystemExit):
        _ = _parse_cone_args([
            'some_file.py',
            '--min_radius', 'notafloat'
        ])
    assert ('--min_radius: invalid float value'
            in errstr(capsys))
    with pytest.raises(SystemExit):
        _ = _parse_cone_args([
            'some_file.py',
            '--max_radius', 'notafloat'
        ])
    assert ('--max_radius: invalid float value'
            in errstr(capsys))



import pytest
from servicemon.query_runner import QueryRunner

from servicemon.query_runner import (
    _parse_query, _parse_replay,
    conegen_defaults, conelist_defaults
)


def errstr(capsys):
    """
    A shortcut with a side effect:  Resets captured stdout and stderr.
    """
    captured = capsys.readouterr()
    result = str(captured.err)
    return result


def test_global_query_args(capsys):

    # With defaults.
    args = _parse_query([
        'my_services_file',
        '--cone_file', 'my_cones.py'
    ])
    assert vars(args) == {'cone_file': 'my_cones.py',
                          'cone_limit': conelist_defaults['cone_limit'],
                          'load_plugins': None,
                          'max_radius': conegen_defaults['max_radius'],
                          'min_radius': conegen_defaults['min_radius'],
                          'norun': False,
                          'num_cones': None,
                          'result_dir': 'results',
                          'save_results': False,
                          'services': 'my_services_file',
                          'start_index': conelist_defaults['start_index'],
                          'tap_mode': 'async',
                          'user_agent': None,
                          'verbose': False,
                          'writers': ['csv_writer']}

    # Without defaults for auto-generated cones.
    custom_agent = 'MyCustomAgent/6.7'
    args = _parse_query([
        'my_services_file',
        '--load_plugins', 'my_plugin_dir',
        '--writer', 'my_writer1', '--writer', 'my_writer2:karg1=val1,kwarg2=val2',
        '--result_dir', 'my_output_dir',
        '--save_results', '--tap_mode', 'sync', '--norun', '--verbose',
        '--user_agent', custom_agent,
        '--num_cones', '22', '--min_radius', '0.123', '--max_radius', '0.456',
        '--start_index', '17', '--cone_limit', '3'
    ])
    assert vars(args) == {'cone_file': None,
                          'cone_limit': 3,
                          'load_plugins': 'my_plugin_dir',
                          'max_radius': 0.456,
                          'min_radius': 0.123,
                          'norun': True,
                          'num_cones': 22,
                          'result_dir': 'my_output_dir',
                          'save_results': True,
                          'services': 'my_services_file',
                          'start_index': 17,
                          'tap_mode': 'sync',
                          'user_agent': custom_agent,
                          'verbose': True,
                          'writers': ['my_writer1', 'my_writer2:karg1=val1,kwarg2=val2']}

    # Without defatults for cone_file specified.
    args = _parse_query([
        'my_services_file',
        '--result_dir', 'my_output_dir',
        '--load_plugins', 'my_plugin_dir',
        '--writer', 'my_writer1', '--writer', 'my_writer2:karg1=val1,kwarg2=val2',
        '--save_results', '--tap_mode', 'sync', '--norun', '--verbose',
        '--cone_file', 'fun_cone_file.py',
        '--start_index', '13', '--cone_limit', '44'
    ])
    assert vars(args) == {'cone_file': 'fun_cone_file.py',
                          'cone_limit': 44,
                          'load_plugins': 'my_plugin_dir',
                          'max_radius': conegen_defaults['max_radius'],
                          'min_radius': conegen_defaults['min_radius'],
                          'norun': True,
                          'num_cones': None,
                          'result_dir': 'my_output_dir',
                          'save_results': True,
                          'services': 'my_services_file',
                          'start_index': 13,
                          'tap_mode': 'sync',
                          'user_agent': None,
                          'verbose': True,
                          'writers': ['my_writer1', 'my_writer2:karg1=val1,kwarg2=val2']}

    # With short args
    args = _parse_query([
        'my_services_file',
        '-l', 'my_plugin_dir',
        '-w', 'my_writer1', '--writer', 'my_writer2:karg1=val1,kwarg2=val2',
        '-s', '-t', 'sync', '-n', '-v',
        '--cone_file', 'fun_cone_file.py',
        '--start_index', '13', '--cone_limit', '44'
    ])
    assert vars(args) == {'cone_file': 'fun_cone_file.py',
                          'cone_limit': 44,
                          'load_plugins': 'my_plugin_dir',
                          'max_radius': conegen_defaults['max_radius'],
                          'min_radius': conegen_defaults['min_radius'],
                          'norun': True,
                          'num_cones': None,
                          'result_dir': 'results',
                          'save_results': True,
                          'services': 'my_services_file',
                          'start_index': 13,
                          'tap_mode': 'sync',
                          'user_agent': None,
                          'verbose': True,
                          'writers': ['my_writer1', 'my_writer2:karg1=val1,kwarg2=val2']}


def test_global_query_arg_errors(capsys):
    with pytest.raises(SystemExit):
        _ = _parse_query([])
    assert ('error: the following arguments are required: service'
            in errstr(capsys))

    with pytest.raises(SystemExit):
        _ = _parse_query(['service_file_only.py'])
    assert ('Either --num-cones or --cone_file must be present'
            in errstr(capsys))

    with pytest.raises(SystemExit):
        _ = _parse_query(['service_file.py',
                          '--min_radius', '0.123',
                          '--max_radius', '0.456'])
    assert ('argument --num-cones is required when --min-radius or --max-radius are present.'
            in errstr(capsys))

    with pytest.raises(SystemExit):
        _ = _parse_query(['service_file.py',
                          '--max_radius', '0.456'])
    assert ('argument --num-cones is required when --min-radius or --max-radius are present.'
            in errstr(capsys))

    # No short arg for cone_file.
    with pytest.raises(SystemExit):
        _ = _parse_query(['service_file.py',
                          '-c', 'some_cone_file.py'])
    assert ('unrecognized arguments: -c'
            in errstr(capsys))


def test_global_replay_args(capsys):

    # With defaults.
    args = _parse_replay([
        'file_to_replay.csv'
    ])
    assert vars(args) == {'cone_limit': 100000000,
                          'file_to_replay': 'file_to_replay.csv',
                          'load_plugins': None,
                          'norun': False,
                          'result_dir': 'results',
                          'save_results': False,
                          'start_index': 0,
                          'tap_mode': 'async',
                          'user_agent': None,
                          'verbose': False,
                          'writers': ['csv_writer']}

    # Without defaults.
    custom_agent = 'AnotherCustomAgent/12.9 (with a comment)'
    args = _parse_replay([
        'file_to_replay.csv',
        '--load_plugins', 'my_plugin_dir',
        '--writer', 'my_writer1', '--writer', 'my_writer2:karg1=val1,kwarg2=val2',
        '--result_dir', 'my_output_dir',
        '--save_results', '--tap_mode', 'sync', '--norun', '--verbose',
        '--user_agent', custom_agent,
        '--start_index', '17', '--cone_limit', '3'
    ])
    assert vars(args) == {'cone_limit': 3,
                          'file_to_replay': 'file_to_replay.csv',
                          'load_plugins': 'my_plugin_dir',
                          'norun': True,
                          'result_dir': 'my_output_dir',
                          'save_results': True,
                          'start_index': 17,
                          'tap_mode': 'sync',
                          'user_agent': custom_agent,
                          'verbose': True,
                          'writers': ['my_writer1', 'my_writer2:karg1=val1,kwarg2=val2']}


def test_global_replay_arg_errors(capsys):

    # cone_file is not available for replay.
    with pytest.raises(SystemExit):
        _ = _parse_replay(['file_to_replay.csv',
                          '--cone_file', 'some_cone_file.py'])
    assert ('unrecognized arguments: --cone_file'
            in errstr(capsys))


def test_service_validation(capsys):
    args = _parse_query([
        'fake_services_file',
        '--cone_file', 'my_cones.py'
    ])
    args.services = []
    args.cone_file = []
    qr = QueryRunner(args)
    with pytest.warns(UserWarning, match='Service list is empty') as record:
        qr._validate_services(args.services)
    assert len(record) == 1

    args.services = [
        {'service_type': 'tap'},
        {'service_type': 'tap'},
        {'service_type': 'tap'},
    ]
    with pytest.warns(None) as record:
        qr._validate_services(args.services)
    assert len(record) == 0

    args.services = [
        {'service_type': 'tap'},
        {'service_type': 'cone'},
        {'service_type': 'tap'},
    ]
    with pytest.warns(UserWarning,
                      match='Differing service_type values found in service list.'
                      '  Some result writers may fail.') as record:
        qr._validate_services(args.services)
    assert len(record) == 1



import time
from datetime import datetime


class Interval():
    """
    """
    def __init__(self, desc, duration=0):
        self._desc = desc
        self._start_time = time.time()
        self._end_time = self._start_time + duration

    def close(self):
        self._end_time = time.time()
        return self

    @property
    def desc(self):
        return self._desc

    @property
    def duration(self):
        return self._end_time - self._start_time

    @property
    def start_time(self):
        return self._start_time

    @property
    def end_time(self):
        return self._end_time


class QueryStats():
    """
    """
    def __init__(self, name, base_name, service_type, access_url, query_params,
                 result_meta_fields, max_intervals=8):

        # First save the params needed to define the result structure.
        self._query_params = self._organize_params(query_params)
        self._result_meta_fields = result_meta_fields
        self._max_intervals = max_intervals
        self._result_meta = dict.fromkeys(result_meta_fields)

        self._vals = dict.fromkeys(self.columns())

        self._vals['name'] = name
        self._vals['base_name'] = base_name
        self._vals['service_type'] = service_type
        self._vals['access_url'] = access_url
        self._vals['errmsg'] = ':'

        self._query_params = self._organize_params(query_params)
        self._vals.update(self._query_params)  # Add the query_params values.
        self._vals.update(self._result_meta)  # Add the result metadata values.

        self._intervals = []

    def add_interval(self, interval):
        lint = len(self._intervals)
        if lint > self._max_intervals:
            raise ValueError(
                f'Too many intervals added ({self._max_intervals + 1})')

        if lint == 0:
            now = datetime.fromtimestamp(time.time())  # now = datetime.now()
            self._vals['start_time'] = now.strftime('%Y-%m-%d %H:%M:%S.%f')
        end = datetime.fromtimestamp(interval.end_time)
        self._vals['end_time'] = end.strftime('%Y-%m-%d %H:%M:%S.%f')

        self._vals[f'int{lint}_desc'] = interval.desc
        self._vals[f'int{lint}_duration'] = interval.duration

        self._intervals.append(interval)

    # property result metadata
    @property
    def result_meta(self):
        return self._result_meta

    @property
    def errmsg(self):
        return self._vals['errmsg']

    @errmsg.setter
    def errmsg(self, val):
        self._vals['errmsg'] = val

    @result_meta.setter
    def result_meta(self, value):
        self._result_meta = value
        for key in self._result_meta_fields:
            self._vals[key] = value.get(key)

    def columns(self):
        cols = ['name', 'start_time', 'end_time']
        for i in range(0, self._max_intervals):
            cols.append(f'int{i}_desc')
            cols.append(f'int{i}_duration')
        cols.append('base_name')
        cols.append('service_type')
        cols.append('RA')
        cols.append('DEC')
        cols.append('SR')
        cols.append('ADQL')
        cols.append('other_params')
        cols.append('access_url')
        cols.append('errmsg')
        cols.extend(list(self.result_meta.keys()))
        return cols

    def row_values(self):
        return self._vals

    def _organize_params(self, in_p):
        fixed_keys = ('RA', 'DEC', 'SR', 'ADQL')
        p = dict.fromkeys(fixed_keys, '')
        p['other_params'] = {}
        for key in in_p.keys():
            if key in fixed_keys:
                p[key] = in_p[key]
            else:
                p['other_params'][key] = in_p[key]
        return p



import os
import sys
import ast
import pprint
from argparse import ArgumentParser

import numpy as np
from numpy.random import random_sample as rand
from astropy import units as u
from astropy.coordinates import SkyCoord


class Cone:
    """
    """

    def __init__(self):
        """
        Not intended to be instantiated.
        """

    @staticmethod
    def random_skycoord():
        """
        """
        ra_rad = (2 * np.pi * rand()) * u.rad
        dec_rad = np.arcsin(2. * (rand() - 0.5)) * u.rad

        skycoord = SkyCoord(ra_rad, dec_rad)
        return skycoord

    @staticmethod
    def random_coords():
        """
        """
        ra_rad = (2 * np.pi * rand()) * u.rad
        ra_deg = ra_rad.to_value(u.deg)
        dec_rad = np.arcsin(2. * (rand() - 0.5)) * u.rad
        dec_deg = dec_rad.to_value(u.deg)

        return {'ra': ra_deg, 'dec': dec_deg}

    @staticmethod
    def random_cone(min_radius, max_radius):
        """
        """
        if not 0 <= min_radius < max_radius:
            raise ValueError('min-radius must be in the range [0,max_radius).')
        coords = Cone.random_coords()
        coords['radius'] = (max_radius - min_radius) * rand() + min_radius
        return coords

    @staticmethod
    def generate_random(num_points, min_radius, max_radius):
        """
        Yield objects with random (and legal) ra, dec, and radius attirbutes.
        """
        if not 0 <= min_radius < max_radius:
            raise ValueError('min-radius must be in the range [0,max_radius).')
        if num_points <= 0:
            raise ValueError('num_points must be a positive number.')

        def cones(num_points, min_radius, max_radius):
            # pylint: disable=unused-variable
            for i in range(num_points):
                cone = Cone.random_cone(min_radius, max_radius)
                yield cone

        return cones(num_points, min_radius, max_radius)

    @staticmethod
    def write_random(num_points, min_radius, max_radius, filename=None):
        generator = Cone.generate_random(num_points, min_radius, max_radius)
        Cone.write_cones(generator, filename)

    @staticmethod
    def write_cones(cones, filename=None):
        stream = sys.stdout
        if filename is not None:
            dirname = os.path.dirname(filename)
            if dirname != '':
                os.makedirs(dirname, exist_ok=True)
            stream = open(filename, "w", encoding="utf-8")

        pp = pprint.PrettyPrinter(width=100, stream=stream, compact=True)
        stream.write("[")
        sep = ''
        indent = '    '
        for cone in cones:
            s = pp.pformat(cone)
            stream.write(f'{sep}\n{indent}{s}')
            sep = ','
        stream.write("\n]\n")

        if filename is not None:
            stream.close()

    @staticmethod
    def reverse_cone_file(infile, outfile=None):
        # read as Python literal
        with open(infile, 'r') as f:
            cones = ast.literal_eval(f.read())

        cones.reverse()

        Cone.write_cones(cones, outfile)


conegen_defaults = {
    'min_radius': 0,
    'max_radius': 0.25
}


def _apply_query_defaults(parsed_args, defaults):
    for k in defaults:
        if getattr(parsed_args, k) is None:
            setattr(parsed_args, k, defaults[k])


def sm_conegen(input_args=None):
    args = _parse_cone_args(input_args)

    Cone.write_random(args.num_cones, args.min_radius, args.max_radius,
                      filename=args.outfile)


def _parse_cone_args(input_args):
    parser = ArgumentParser(description='Generate random cones.')

    parser.add_argument(
        'outfile', help="Name of the output file to contain the cones. ")
    parser.add_argument(
        '--num_cones', type=int, metavar='num_cones', required=True,
        help='Number of cones to generate')
    parser.add_argument(
        '--min_radius', type=float, metavar='min_radius',
        help='Minimum radius (deg).'
        f' Default={conegen_defaults["min_radius"]}')
    parser.add_argument(
        '--max_radius', type=float, metavar='max_radius',
        help='Maximum radius (deg).'
        f' Default={conegen_defaults["max_radius"]}')

    args = parser.parse_args(input_args)

    _apply_query_defaults(args, conegen_defaults)
    return args



[
    {'base_name': 'SampleSIAService',
    'service_type': 'xcone',
    'adql': '',
    'access_url': 'https://sia.some_archive.org/'
    'mean.votable?flatten_response=false&sort_by=distance'
    '&ra={}&dec={}&radius={}'
    }
]



from servicemon.plugin_support import AbstractTimedQuery


class UserQuery2(AbstractTimedQuery, plugin_name='userquery2',
                 description="Another query plug-in in the user's dir that does nothing."):
    pass



[
    {'base_name': 'PanSTARRS',
     'service_type': 'xcone',
     'adql': '',
     'access_url': 'https://catalogs.mast.stsci.edu/api/v0.1/panstarrs/dr2/'
     'mean.votable?flatten_response=false&raw=false&sort_by=distance'
     '&ra={}&dec={}&radius={}'
     }
]



[
    {'base_name': 'NED',
     'service_type': 'tap',
     'access_url': 'https://ned.ipac.caltech.edu/tap/',
     'adql':'''
SELECT
   *
   FROM NEDTAP.objdir
   WHERE
      1=CONTAINS(POINT('ICRS', ra, dec),
                 CIRCLE('ICRS', {}, {}, {} ))
     '''
     }
]



import os
import shutil
import pytest
from pathlib import Path
from datetime import datetime

from astropy.table import Table
from servicemon.plugin_support import SmPluginSupport, AbstractResultWriter
from servicemon.query_runner import _parse_query


def errstr(capsys):
    """
    A shortcut with a side effect:  Resets captured stdout and stderr.
    """
    captured = capsys.readouterr()
    result_err = str(captured.err)
    result_out = str(captured.out)
    return result_err, result_out


@pytest.fixture
def cleandirs():
    # Clean before
    shutil.rmtree(Path('my_fake_result_dir'), ignore_errors=True)
    shutil.rmtree(Path('my_fake_result_dir2'), ignore_errors=True)
    yield
    # And after
    shutil.rmtree(Path('my_fake_result_dir'), ignore_errors=True)
    shutil.rmtree(Path('my_fake_result_dir2'), ignore_errors=True)


def test_end(capsys):
    SmPluginSupport.load_builtin_plugins()

    # Test with no input args.
    cw = begin_no_outfile()
    cw.end()


@pytest.mark.usefixtures("cleandirs")
def test_begin(capsys):
    SmPluginSupport.load_builtin_plugins()
    now = datetime.now()
    dtstr = now.strftime('%Y-%m-%d')

    # Test with no input args.
    cw = begin_no_outfile()
    assert isinstance(cw._outfile_path, Path)
    assert cw._outfile_path.stem.startswith(f'fake_services_file_{dtstr}')
    assert cw._outfile_path.parent.name == 'my_fake_result_dir'
    os.rmdir('my_fake_result_dir')

    # Test with outfile specified with fixed dstr.  Dstring substitutions are not done at this level.
    cw = begin_w_outfile(dtstr)
    assert isinstance(cw._outfile_path, Path)
    assert cw._outfile_path.name == f'myout-{dtstr}.csv'
    assert cw._outfile_path.parent.name == 'my_fake_result_dir2'
    os.rmdir('my_fake_result_dir2')

    # Test with outfile as stdout
    cw = begin_outfile_stdout()
    assert cw._outfile_path is None
    assert not Path('my_fake_result_dir3').exists()


@pytest.mark.usefixtures("cleandirs")
def test_one_result(capsys):
    SmPluginSupport.load_builtin_plugins()

    # Test with specified outfile.
    cw = begin_w_outfile('nodate')
    resfile = write_some_results(cw)

    # Read and check results.
    t = Table.read(resfile, format='csv')
    assert_table_vals(t)

    # Test with stdout.
    cw = begin_outfile_stdout()
    resfile = write_some_results(cw)
    _, out = errstr(capsys)
    t = Table.read(out, format='csv')
    assert_table_vals(t)


def assert_table_vals(t):
    assert len(t) == 2
    assert t['a'][0] == '1'
    assert t['a'][1] == 'vala'
    assert t['b'][0] == '2'
    assert t['b'][1] == 'valb'
    assert t['c'][0] == '3'
    assert t['c'][1] == 'valc'


def write_some_results(cw):
    resfile = Path('my_fake_result_dir2') / Path('myout-nodate.csv')
    results = Results(['a', 'b', 'c'])
    results.set_values({'a': 1, 'b': 2, 'c': 3})
    cw.one_result(results)
    results.set_values({'a': 'vala', 'b': 'valb', 'c': 'valc'})
    cw.one_result(results)
    return resfile


def begin_outfile_stdout():
    args = _parse_query([
        'fake_services_file.py',
        '--cone_file', 'my_cones.py',
        '--result_dir', 'my_fake_result_dir3'
    ])
    cw_plugin = AbstractResultWriter.get_plugin_from_spec('csv_writer:outfile=stdout')
    cw = cw_plugin.cls()
    cw.begin(args, **cw_plugin.kwargs)
    return cw


def begin_w_outfile(dtstr):
    args = _parse_query([
        'fake_services_file.py',
        '--cone_file', 'my_cones.py',
        '--result_dir', 'my_fake_result_dir2'
    ])
    cw_plugin = AbstractResultWriter.get_plugin_from_spec(f'csv_writer:outfile=my_fake_result_dir2/myout-{dtstr}.csv')
    cw = cw_plugin.cls()
    cw.begin(args, **cw_plugin.kwargs)
    return cw


def begin_no_outfile():
    args = _parse_query([
        'fake_services_file.py',
        '--cone_file', 'my_cones.py',
        '--result_dir', 'my_fake_result_dir'
    ])
    cw_plugin = AbstractResultWriter.get_plugin_from_spec('csv_writer')
    cw = cw_plugin.cls()
    cw.begin(args, **cw_plugin.kwargs)
    return cw


class Results():
    def __init__(self, cols):
        self._cols = cols
        self._vals = [None] * len(self._cols)

    def set_values(self, vals):
        if len(vals) != len(self._cols):
            raise ValueError('values must have same length as columns.')
        self._vals = vals

    def columns(self):
        return self._cols

    def row_values(self):
        return self._vals



# import os

# If this package has tests data in the tests/data directory, add them to
# the paths here, see commented example
paths = ['coveragerc',
         # os.path.join('data', '*fits')
         ]


def get_package_data():
    return {
        _ASTROPY_PACKAGE_NAME_ + '.tests': paths} # noqa



[
    {'base_name': 'IPAC_2MASS',
     'service_type': 'tap',
     'access_url': 'https://irsa.ipac.caltech.edu/TAP',
     'adql':'''
    SELECT ra, dec, j_m, h_m, k_m FROM fp_psc
    WHERE CONTAINS(POINT('ICRS',ra, dec),
    CIRCLE('J2000',{},{},{}))=1
     '''
     }
]



import requests
import time
from time import sleep
from distutils.version import LooseVersion
from functools import partial

from codetiming import Timer
from astropy.io.votable import parse as votableparse

from pyvo.dal.tap import AsyncTAPJob, TAPService, TAPQuery, TAPResults
from pyvo.dal.exceptions import DALQueryError, DALServiceError
from pyvo.utils.http import use_session
from pyvo.io import uws


class TAPServiceSM(TAPService):
    """
    Specialization of TAPService to support timing.
    """

    def run_sync_timed(
            self, query, language="ADQL", maxrec=None, uploads=None,
            streamable_response=False,
            **keywords):
        """
        runs sync query and returns its result

        Parameters
        ----------
        query : str
            The query
        language : str
            specifies the query language, default ADQL.
            useful for services which allow to use the backend query language.
        maxrec : int
            the maximum records to return. defaults to the service default
        uploads : dict
            a mapping from table names to objects containing a votable
        streamable_response: bool
            If False (default) return TapResult, otherwise return
            a streamable response.

        Returns
        -------
        TAPResults
            the query result

        See Also
        --------
        TAPResults
        """
        tap_query = self.create_query(
            query, language=language, maxrec=maxrec, uploads=uploads,
            **keywords)

        if streamable_response:
            # This block adapted from DALQuery.execute_stream()
            # to get a response instead of a response.raw.

            result = tap_query.submit()

            # Not clear if we want to catch this for the servicemon use case.
            try:
                result.raise_for_status()
            except requests.RequestException as ex:
                raise DALServiceError.from_except(ex)

        else:
            result = tap_query.execute()

        return result

    def run_async_timed(
            self, query, language="ADQL", maxrec=None, uploads=None,
            streamable_response=False, delete=False, **keywords):
        """
        runs async query and returns its result
         Parameters
        ----------
        query : str, dict
            the query string / parameters
        language : str
            specifies the query language, default ADQL.
            useful for services which allow to use the backend query language.
        maxrec : int
            the maximum records to return. defaults to the service default
        uploads : dict
            a mapping from table names to objects containing a votable
        streamable_response: bool
            If False (default) return TapResult, otherwise return
            a streamable response.
         Returns
        -------
        TAPResult or requests.packages.urllib3.response.HTTPResponse
            the query instance
         Raises
        ------
        DALServiceError
           for errors connecting to or communicating with the service
        DALQueryError
           for errors either in the input query syntax or
           other user errors detected by the service
        DALFormatError
           for errors parsing the VOTable response
         See Also
        --------
        AsyncTAPJob
        """
        with Timer('submit', logger=None):
            job = AsyncTAPSM.create(
                self.baseurl, query, language, maxrec, uploads, self._session, **keywords)

        with Timer('run', logger=None):
            job = job.run()

        with Timer('wait', logger=None):
            job = job.wait()

        with Timer('raise_if_error', logger=None):
            if job._job.phase in {"ERROR", "ABORTED"}:
                raise DALQueryError("Query Error", job._job.phase, job.url)

        with Timer('fetch_response', logger=None):
            if streamable_response:
                result = job.get_result_response()
            else:
                result = job.fetch_result()

        if delete:
            with Timer('delete', logger=None):
                job.delete()

        return result


class AsyncTAPSM(AsyncTAPJob):
    """
    Specialization of pyvo AsyncTAPJob to support timing.
    """

    _job = {}

    @classmethod
    def create(
            cls, baseurl, query, language="ADQL", maxrec=None, uploads=None,
            session=None, **keywords):
        """
        creates a async tap job on the server under `baseurl`
        Parameters
        ----------
        baseurl : str
            the TAP baseurl
        query : str
            the query string
        language : str
            specifies the query language, default ADQL.
            useful for services which allow to use the backend query language.
        maxrec : int
            the maximum records to return. defaults to the service default
        uploads : dict
            a mapping from table names to objects containing a votable
        session : object
           optional session to use for network requests
        """
        query = TAPQuery(
            baseurl, query, mode="async", language=language, maxrec=maxrec,
            uploads=uploads, session=session, **keywords)
        response = query.submit()
        uws_job = uws.parse_job(response.raw.read)
        job = cls(response.url, uws_job=uws_job, session=session)
        return job

    def __init__(self, url, uws_job=None, session=None):
        """
        initialize the job object with the given url and fetch remote values
        Parameters
        ----------
        url : str
            the job url
        """
        self._url = url
        self._session = use_session(session)

        self._job = uws_job

    def _update(self, wait_for_statechange=False, timeout=10.):
        """
        updates local job infos with remote values
        """
        try:
            if wait_for_statechange:
                response = self._session.get(
                    self.url, stream=True, timeout=timeout+5, params={
                        "WAIT": str(timeout)
                    }
                )
            else:
                response = self._session.get(self.url, stream=True, timeout=timeout)
            response.raise_for_status()
        except requests.RequestException as ex:
            raise DALServiceError.from_except(ex, self.url)

        # requests doesn't decode the content by default
        response.raw.read = partial(response.raw.read, decode_content=True)

        self._job = uws.parse_job(response.raw.read)

    def wait(self, phases=None, async_request_timeout=30,
             async_total_timeout=120):
        """
        waits for the job to reach the given phases.
        Parameters
        ----------
        phases : list
            phases to wait for
        Raises
        ------
        DALServiceError
            if the job is in a state that won't lead to an result
        """
        if not phases:
            phases = {"COMPLETED", "ABORTED", "ERROR"}

        interval = 1.0
        increment = 1.2

        active_phases = {
            "QUEUED", "EXECUTING", "RUN", "COMPLETED", "ERROR", "UNKNOWN"}

        supports_wait_for_statechange = (LooseVersion(self._job.version) >=
                                         LooseVersion("1.1"))
        if supports_wait_for_statechange:
            timeout = async_total_timeout
        else:
            timeout = async_request_timeout
        start_time = time.time()
        while True:
            self._update(wait_for_statechange=supports_wait_for_statechange,
                         timeout=timeout)

            elapsed = time.time() - start_time

            # use the cached value
            cur_phase = self._job.phase

            if cur_phase not in active_phases:
                raise DALServiceError(
                    "Cannot wait for job completion. Job is not active!")

            if cur_phase in phases:
                break

            # Check if we've exceeded total_timeout
            if elapsed > async_total_timeout:
                raise DALServiceError(
                    f'Async TAP job timed out, exceeding {async_total_timeout}s.')

            # fallback for uws 1.0
            if not supports_wait_for_statechange:
                sleep(interval)
                interval = min(120, interval * increment)

        return self

    def fetch_result(self):
        """
        returns the result votable if query is finished
        """
        response = self.get_result_response()

        response.raw.read = partial(
            response.raw.read, decode_content=True)
        return TAPResults(votableparse(response.raw.read), url=self.result_uri, session=self._session)

    def get_result_response(self):
        try:
            response = self._session.get(self.result_uri, stream=True)
            response.raise_for_status()
        except requests.RequestException as ex:
            # Got rid of an _update() that was here because it does yet another query.
            # It was probably there to ensure that _phase is set, but that's hopeless if we're
            # getting errors here.

            # We can add specialized error handling here if needed.
            self.raise_if_error()
            raise DALServiceError.from_except(ex, self.url)

        return response



import time
from servicemon.query_stats import Interval, QueryStats


def test_interval():
    i = Interval('MyInterval')
    time.sleep(0.001)
    i.close()
    assert i.desc == 'MyInterval'
    assert 0.0 < i.duration < 0.1


def test_params():
    qs = QueryStats('HSC_cone_123.4_56.7_8.9', 'HSC', 'cone',
                    'http://google.com',
                    {'RA': 123.4, 'DEC': 56.7, 'SR': 8.9},
                    {'meta1': 14, 'meta2': 'important data'})
    rv = qs.row_values()
    assert rv['RA'] == 123.4
    assert rv['DEC'] == 56.7
    assert rv['SR'] == 8.9
    assert rv['ADQL'] == ''
    assert rv['other_params'] == {}
    assert 'int0_desc' in qs.columns()
    assert 'int0_duration' in qs.columns()
    assert 'int1_desc' in qs.columns()
    assert 'int1_duration' in qs.columns()
    assert 'int8_desc' not in qs.columns()
    assert 'meta1' in qs.columns()
    assert 'meta2' in qs.columns()

    qs = QueryStats('Pan-STARRS_random_tap', 'Pan-STARRS_', 'tap',
                    'http://google.com',
                    {'ADQL': 'select cool_stuff from some_table'},
                    {'meta1': 14, 'meta2': 'important data', 'meta3': 42},
                    max_intervals=3)
    rv = qs.row_values()
    assert rv['RA'] == ''
    assert rv['DEC'] == ''
    assert rv['SR'] == ''
    assert rv['ADQL'] == 'select cool_stuff from some_table'
    assert rv['other_params'] == {}
    assert 'int2_desc' in qs.columns()
    assert 'int2_duration' in qs.columns()
    assert 'meta3' in qs.columns()

    qs = QueryStats('HSC_cone_123.4_56.7_8.9', 'HSC', 'cone',
                    'http://google.com',
                    {'RA': 123.4, 'DEC': 56.7, 'SR': 8.9, 'verbose': 3},
                    {'meta1': 14, 'meta2': 'important data'})
    rv = qs.row_values()
    assert rv['other_params']['verbose'] == 3



[
    # This one uses arcminutes as radius for cone according to docs on
    # http://cxc.harvard.edu/csc2/cli/index.html#vo
    {'base_name': 'CSC',
     'service_type': 'tap',
     'access_url': 'http://cda.harvard.edu/csctap',
     'adql': '''
     SELECT m.name, m.ra, m.dec, m.flux_aper_b FROM csc2.master_source m
     WHERE csc2.cone_distance(m.ra,m.dec,{},{})<={}
     '''
     }
]



from servicemon.plugin_support import AbstractResultWriter


class BasicWriter(AbstractResultWriter, plugin_name='basic_writer',
                  description='Another writer in the default plug-in dir that does nothing.'):

    def begin(self, args, **kwargs):
        pass

    def one_result(self, stats):
        pass

    def end(self):
        pass



import pyvo as vo


class StatQueries():

    @property
    def tap_service(self):
        return self._tap_service

    def __init__(self, tap_url='http://navo01.ipac.caltech.edu/TAP'):
        """
        """
        self._tap_service = vo.dal.TAPService(tap_url)

    def _build_stat_query(self, top=None, base_name=None, service_type=None, start_time=None, end_time=None):
        """
        """
        where_list = []
        if base_name is not None:
            if isinstance(base_name, str):
                where_list.append(f"base_name like '{base_name}'")
            else:
                bases = ", ".join([f"'{b}'" for b in base_name])
                where_list.append(f"base_name in ({bases})")

        if service_type is not None:
            if isinstance(service_type, str):
                where_list.append(f"service_type like '{service_type}'")
            else:
                types = ", ".join([f"'{s}'" for s in service_type])
                where_list.append(f"service_type in ({types})")

        if start_time is not None:
            where_list.append(f"start_time >= '{start_time}'")
        if end_time is not None:
            where_list.append(f"end_time <= '{end_time}'")

        top_clause = ''if top is None else f'top {top} '
        where_clause = ''
        if where_list:
            where_clause = "\nwhere\n" + "\nand ".join(where_list)
        query = f"select {top_clause}* from navostats{where_clause}"

        return query

    def do_query(self, query):
        results = self._tap_service.search(query)
        data_table = results.to_table()

        return data_table

    def do_stat_query(self, top=None, base_name=None, service_type=None, start_time=None, end_time=None):
        """
        """
        query = self._build_stat_query(top, base_name, service_type, start_time, end_time)
        data_table = self.do_query(query)

        return data_table

    def get_name_service_pairs(self):
        """
        Queries the DB to find all unique (base_name, service_type) pairs and
        returns the results as a list of tuples.
        """

        query = """
        select base_name, service_type from navostats
        group by base_name, service_type
        order by base_name
        """
        data_table = self.do_query(query)

        name_service_pairs = [(r['base_name'], r['service_type']) for r in data_table]
        return name_service_pairs



import os
import json
import socket
import requests
import datetime

from ec2_metadata import ec2_metadata

from servicemon.plugin_support import AbstractResultWriter


class AWSResultWriter(AbstractResultWriter, plugin_name='aws_writer',
                      description='Sends results to a central SQLite database.'):

    # BEGIN gathers info and sends a 'begin' state message
    # to the admin database

    def begin(self, args, outfile=None):

        self._pid = os.getpid()

        self._starttime = str(datetime.datetime.now())

        self._adminURL = 'http://vmnavo01.ipac.caltech.edu/cgi-bin/NAVOMonitor/nph-admin'
        self._resultsURL = 'http://vmnavo01.ipac.caltech.edu/cgi-bin/NAVOMonitor/nph-submit'

        try:
            self._region = ec2_metadata.region
            self._instance_id = ec2_metadata.instance_id
            self._hostname = ec2_metadata.public_hostname

        except Exception:
            self._region = ''
            self._instance_id = ''
            self._hostname = socket.getfqdn()

        if len(self._region) > 0:
            self._location = self._region
        else:
            self._location = self._hostname

        data = {'hostname': self._hostname, 'pid': self._pid,
                'region': self._region, 'instance_id': self._instance_id,
                'starttime': self._starttime, 'endtime': ''}

        jsonStr = json.dumps(data)

        print("START> ", jsonStr)

        post_data = {'json': jsonStr}

        r = requests.post(url=self._adminURL, files=post_data)

        print(r.text)

    # END sends an 'end' state message to the admin database,
    # potentially triggering shutdown/clean-up

    def end(self):

        endtime = str(datetime.datetime.now())

        data = {'hostname': self._hostname, 'pid': self._pid,
                'region': self._region, 'instance_id': self._instance_id,
                'starttime': self._starttime, 'endtime': endtime}

        jsonStr = json.dumps(data)

        print("END>   ", jsonStr)

        post_data = {'json': jsonStr}

        r = requests.post(url=self._adminURL, files=post_data)

        print(r.text)

    # ONE_RESULT sends a data record (the results on a single query)
    # to the results database

    def one_result(self, stats):

        fields = stats.columns()

        self.ncols = len(fields)

        row = stats.row_values()

        row['location'] = self._location

        print('DATA> ' + str(row))

        jsonStr = json.dumps(row)

        post_data = {'json': jsonStr}

        r = requests.post(url=self._resultsURL, files=post_data)

        print(r.text)



from servicemon.analysis.stat_queries import StatQueries


def test_build_queries(capsys):
    sq = StatQueries()

    q = sq._build_stat_query()
    assert q == "select * from navostats"

    q = sq._build_stat_query(top=25)
    assert q == "select top 25 * from navostats"

    q = sq._build_stat_query(top=25, base_name='IPAC_%')
    assert q == """select top 25 * from navostats
where
base_name like 'IPAC_%'"""

    q = sq._build_stat_query(top=25, base_name=['IPAC_2MASS', 'GAVO_2MASS'])
    assert q == """select top 25 * from navostats
where
base_name in ('IPAC_2MASS', 'GAVO_2MASS')"""

    q = sq._build_stat_query(top=25, base_name=['IPAC_2MASS', 'GAVO_2MASS'],
                             service_type='tap')
    assert q == """select top 25 * from navostats
where
base_name in ('IPAC_2MASS', 'GAVO_2MASS')
and service_type like 'tap'"""

    q = sq._build_stat_query(top=25, base_name=['IPAC_2MASS', 'GAVO_2MASS'],
                             service_type=['tap', 'cone'])
    assert q == """select top 25 * from navostats
where
base_name in ('IPAC_2MASS', 'GAVO_2MASS')
and service_type in ('tap', 'cone')"""

    q = sq._build_stat_query(top=25, service_type=['tap', 'cone'])
    assert q == """select top 25 * from navostats
where
service_type in ('tap', 'cone')"""

    q = sq._build_stat_query(service_type='cone', start_time='2021-02-01')
    assert q == """select * from navostats
where
service_type like 'cone'
and start_time >= '2021-02-01'"""

    q = sq._build_stat_query(service_type='cone', start_time='2021-02-01',
                             end_time='2021-02-28')
    assert q == """select * from navostats
where
service_type like 'cone'
and start_time >= '2021-02-01'
and end_time <= '2021-02-28'"""



[
    {'base_name': 'STScI_ObsTAP',
     'service_type': 'tap',
     'access_url': 'http://vao.stsci.edu/CAOMTAP/TapService.aspx',
     'adql':'''
    SELECT *
    FROM ivoa.obscore
    WHERE
      1=CONTAINS(POINT('ICRS', s_ra, s_dec),
                 CIRCLE('ICRS', {}, {}, {} ))
     '''
     }
]



import sys
import ast
import csv
import signal
import faulthandler
import platform
import logging
import warnings

from argparse import ArgumentParser
from datetime import datetime

from astropy.table import Table
from .query import Query
from .cone import Cone
from .plugin_support import SmPluginSupport, AbstractResultWriter


class QueryRunner():
    """
    """

    _first_instance = True

    def __init__(self, args):
        """
        args is assumed to be a Namespace object args with each attribute having a valid
        value (no defaults will be applied here).  Using vars(args) to show the object as
        a dict would yield an object like this:

        {'cone_file': 'input/cones-10000-0_05-0_25.py',
         'cone_limit': 2,
         'load_plugins': None,
         'max_radius': 0.25,
         'min_radius': 0,
         'norun': True,
         'num_cones': None,
         'result_dir': 'output',
         'save_results': True,
         'services': 'input/stsci/ps_tap.py',
         'start_index': 14,
         'tap_mode': 'async',
         'verbose': False,
         'writers': ['csv_writer:outfile=output/ps-tap-2020-08-02-16:19:37.484191.csv',
         'some_writer']}
        """

        # reset this flag for each instance
        self._first_stat = True

        self._args = args  # All the args rolled up.  We can get rid of the others.

        services = getattr(args, 'services', None)
        if services is not None:
            self._services = self._read_if_file(services)
        else:
            # We must be in replay mode.
            self._services = self._read_if_file(args.file_to_replay)

        # Use getattr for cone_file since it won't exist for replay use cases.
        self._cones = self._read_if_file(getattr(args, 'cone_file', None))

        self._result_dir = args.result_dir
        self._starting_cone = int(args.start_index)
        self._cone_limit = int(args.cone_limit)
        self._tap_mode = args.tap_mode
        self._user_agent = args.user_agent
        self._save_results = args.save_results
        self._verbose = args.verbose
        self._writer_specs = args.writers

        self._writers_descs = []
        self._writers = []
        self._load_plugins(args)

        # This flag will stay False for all future instances.
        self._first_instance = False

    def _load_plugins(self, args):
        if self._first_instance:
            SmPluginSupport.load_builtin_plugins()

            # Load user plugins.
            if args.load_plugins is None:
                SmPluginSupport.load_plugins()
            else:
                SmPluginSupport.load_plugins(plugins=args.load_plugins)

        if self._writer_specs is None:
            # Default to the csv_writer and its default kwargs.
            self._writer_specs = ['csv_writer']

        for spec in self._writer_specs:
            # Substitute time formats in writer templates,
            # such as '%Y-%m-%d %H:%M:%S.%f'
            now = datetime.now()
            spec_with_time = now.strftime(spec)

            plugin = AbstractResultWriter.get_plugin_from_spec(spec_with_time)
            self._writers_descs.append(plugin)

    def run(self):
        """
        """

        self._validate_services(self._services)

        for wdesc in self._writers_descs:
            w = wdesc.cls()
            w.begin(self._args, **wdesc.kwargs)
            self._writers.append(w)

        if self._cones is not None:
            self._run_with_cones()
        else:
            self._run_services_only()

        for w in self._writers:
            w.end()

    def _validate_services(self, services):
        if len(services) == 0:
            warnings.warn('Service list is empty.  Nothing will be timed.')
        else:
            first_stype = self.getval(services[0], 'service_type')
            for service in services[1:]:
                stype = self.getval(service, 'service_type')
                if stype != first_stype:
                    warnings.warn('Differing service_type values found in service list.'
                                  '  Some result writers may fail.')
                    break

    def _run_with_cones(self):
        cone_index = 0
        cones_run = 0
        for cone in self._cones:
            if cone_index >= self._starting_cone:
                cones_run += 1
                if cones_run > self._cone_limit:
                    break
                for service in self._services:
                    # Don't use the previous results upon new exception.
                    query = None
                    try:
                        query = Query(service, (cone['ra'], cone['dec']),
                                      cone['radius'], self._result_dir,
                                      tap_mode=self._tap_mode,
                                      agent=self._user_agent,
                                      save_results=self._save_results,
                                      verbose=self._verbose)
                        query.run()
                    except Exception as e:
                        msg = f'Query error for cone {cone}, service {service}: {repr(e)}'
                        query._handle_exc(msg, trace=True)
                    try:
                        self._collect_stats(query.stats)
                    except Exception as e:
                        msg = f'Unable to write stats for cone {cone}, service {service}: {repr(e)}'
                        query._handle_exc(msg)

            cone_index += 1

    def _run_services_only(self):
        cone_index = 0
        cones_run = 0
        for service in self._services:
            if cone_index >= self._starting_cone:
                cones_run += 1
                if cones_run > self._cone_limit:
                    break
                try:
                    query = Query(service, None, None, self._result_dir,
                                  tap_mode=self._tap_mode,
                                  agent=self._user_agent,
                                  save_results=self._save_results,
                                  verbose=self._verbose)
                    query.run()
                except Exception as e:
                    msg = f'Query error for service {service}: {repr(e)}'
                    query._handle_exc(msg)
                try:
                    self._collect_stats(query.stats)
                except Exception as e:
                    msg = f'Unable to write stats for service {service}: {repr(e)}'
                    query._handle_exc(msg)

            cone_index += 1

    def _collect_stats(self, stats):
        self._output_stats_row(stats)

    def _output_stats_row(self, stats):
        for w in self._writers:
            w.one_result(stats)

    def _output_stats_row_to_file(self, stats, stat_file):
        writer = csv.DictWriter(stat_file, dialect='excel',
                                fieldnames=stats.columns())
        if self._first_stat:
            self._first_stat = False
            writer.writeheader()
        writer.writerow(stats.row_values())

    def _read_if_file(self, obj):
        val = obj
        if isinstance(obj, str):
            # Read from file
            if obj.endswith('.py'):
                # read as Python literal, then into Table
                with open(obj, 'r') as f:
                    data = ast.literal_eval(f.read())
                val = Table(rows=data)
            else:
                # assume csv and read into Table
                val = Table.read(obj, format='ascii.csv')
        return val

    def getval(self, obj, key, default=None):
        """
        Gets the value as either an attribute val or key val.
        """
        val = getattr(obj, key, None)
        if val is None:
            try:
                val = obj[key]
            except KeyError:
                val = default
        return val


################################################################################

# run time routines

conegen_defaults = {
    'min_radius': 0,
    'max_radius': 0.25
}

conelist_defaults = {
    'start_index': 0,
    'cone_limit': 100000000
}


def sm_query(input_args=None):

    args = _parse_query(input_args)

    _init_logging(args)
    _print_arg_info(args)

    # Get the input from the specified file or have it autogenerated.
    if args.cone_file is None:
        args.cone_file = Cone.generate_random(args.num_cones, args.min_radius, args.max_radius)

    # Run the queries
    qr = QueryRunner(args)
    qr.run()


def _parse_query(input_args):
    """
    # Parse args and apply defaults.
    """
    parser = _create_query_argparser()

    # Parse the arguments.  If args is None, then the args implicitly come from sys.argv.
    args = parser.parse_args(input_args)

    # Catch SIGHUP, SIGQUIT and SIGTERM to allow running in the background.
    catch_signals()

    # Validate args.
    if ((args.min_radius is not None or
         args.max_radius is not None) and
            args.num_cones is None):
        parser.error(message='argument --num-cones is required when '
                     '--min-radius or --max-radius are present.')

    if args.num_cones is None and args.cone_file is None:
        parser.error(message='Either --num-cones or --cone_file must be present\n'
                     '   to specify what values go into the service file templates.')

    # Apply defaults that couldn't be built in.
    apply_query_defaults(args, conegen_defaults)
    apply_query_defaults(args, conelist_defaults)
    if args.writers is None:
        # Default to the csv_writer and its default output file.
        args.writers = ['csv_writer']

    return args


def sm_replay(input_args=None):

    args = _parse_replay(input_args)

    _init_logging(args)
    _print_arg_info(args)

    # Run the queries
    qr = QueryRunner(args)
    qr.run()


def _print_arg_info(args):

    # Print arg info and exit if --norun specified
    if args.verbose or args.norun:
        # print arg info
        import pprint
        pp = pprint.PrettyPrinter(width=100, stream=sys.stdout, compact=True)
        arg_values = pp.pformat(vars(args))
        msg = f'Argument values after parsing and applying defaults:\n{arg_values}'

        if args.norun:
            print(msg)
            exit(0)
        else:
            logging.info(msg)


def _parse_replay(input_args):
    """
    # Parse args and apply defaults.
    """
    parser = _create_replay_argparser()

    # Parse the arguments.  If args is None, then the args implicitly come from sys.argv.
    args = parser.parse_args(input_args)

    # Catch SIGHUP, SIGQUIT and SIGTERM to allow running in the background.
    catch_signals()

    # Apply defaults that couldn't be built in.
    apply_query_defaults(args, conelist_defaults)
    if args.writers is None:
        # Default to the csv_writer and its default output file.
        args.writers = ['csv_writer']
    return args


def _create_query_argparser():
    parser = ArgumentParser(description='Measure query performance.')

    # Add positional args.
    parser.add_argument(
        'services', help='File containing list of services to query')

    # Add general args.
    parser.add_argument('-r', '--result_dir', dest='result_dir', default='results',
                        help='The directory in which to put query result files.'
                        ' Unless --save_results is specified, each query result file'
                        ' will be deleted after statistics are gathered for the query.',
                        metavar='result_dir')
    parser.add_argument('-l', '--load_plugins', dest='load_plugins', metavar='plugin_dir_or_file',
                        help='Directory or file from which to load user plug-ins. '
                        'If not specified, and there is a "plugins" subdirectory, plugin '
                        'files will be loaded from there.')
    parser.add_argument('-w', '--writer', dest='writers', action='append',
                        help="Name and kwargs of a writer plug-in to use."
                        "Format is writer_name[:arg1=val1[,arg2=val2...]]"
                        " May appear multiple times to specify multiple writers."
                        " May contain Python datetime format elements which will be"
                        " substituted with appropriate elements of the current time"
                        " (e.g., results-'%%m-%%d-%%H:%%M:%%S'.py)",
                        metavar='writer')
    parser.add_argument('-s', '--save_results', dest='save_results', action='store_true',
                        help='Save the query result data files.  Without this argument, '
                        'the query result file will be deleted after metadata is gathered '
                        'for the query.')
    parser.add_argument('-t', '--tap_mode', dest='tap_mode',
                        choices={'sync', 'async'}, default='async',
                        help='How to run TAP queries (default=async)')
    parser.add_argument('-u', '--user_agent', dest='user_agent',
                        default=None,
                        help='Override the User-Agent used for queries (default=None)')
    parser.add_argument('-n', '--norun', dest='norun', action='store_true',
                        help='Display summary of command arguments without '
                        'performing any actions')
    parser.add_argument('-v', '--verbose', dest='verbose',
                        action='store_true',
                        help='Print additional information to stderr')

    # Add cone arguments.
    cone_types = parser.add_mutually_exclusive_group()
    cone_types.add_argument('--num_cones', type=int, metavar='num_cones',
                            help='Number of cones to generate')
    cone_types.add_argument('--cone_file', metavar='cone_file',
                            help='Path of the file containing the individual query inputs.')

    cone_random = parser.add_argument_group()
    cone_random.add_argument(
        '--min_radius', type=float, help='Minimum radius (deg).'
        f' Default={conegen_defaults["min_radius"]}')
    cone_random.add_argument(
        '--max_radius', type=float, help='Maximum radius (deg).'
        f' Default={conegen_defaults["max_radius"]}')

    cone_file = parser.add_argument_group()
    cone_file.add_argument(
        '--start_index', type=int, metavar='start_index',
        help='Start with this cone in cone file'
        f' Default={conelist_defaults["start_index"]}')
    cone_file.add_argument(
        '--cone_limit', type=int, metavar='cone_limit',
        help='Maximum number of cones to query'
        f' Default={conelist_defaults["cone_limit"]}')

    return parser


def _create_replay_argparser():
    parser = ArgumentParser(description='Measure query replay performance.')

    # Add positional args.
    parser.add_argument('file_to_replay',
                        help='File containing the results of a previous set of query timings.')

    # Add general args.
    parser.add_argument('-r', '--result_dir', dest='result_dir', default='results',
                        help='The directory in which to put query result files.',
                        metavar='result_dir')
    parser.add_argument('-l', '--load_plugins', dest='load_plugins', metavar='plugin_dir_or_file',
                        help='Directory or file from which to load user plug-ins. '
                        'If not specified, and there is a "plugins" subdirectory, plugin '
                        'files will be loaded from there.')
    parser.add_argument('-w', '--writer', dest='writers', action='append',
                        help="Name and kwargs of a writer plug-in to use."
                        "Format is writer_name[:arg1=val1[,arg2=val2...]]"
                        " May appear multiple times to specify multiple writers."
                        " May contain Python datetime format elements which will be"
                        " substituted with appropriate elements of the current time"
                        " (e.g., results-'%%m-%%d-%%H:%%M:%%S'.py)",
                        metavar='writer')
    parser.add_argument('-s', '--save_results', dest='save_results', action='store_true',
                        help='Save the query result data files.  Without this argument, '
                        'the query result file will be deleted after metadata is gathered '
                        'for the query.')
    parser.add_argument('-t', '--tap_mode', dest='tap_mode',
                        choices={'sync', 'async'}, default='async',
                        help='How to run TAP queries (default=async)')
    parser.add_argument('-u', '--user_agent', dest='user_agent',
                        default=None,
                        help='Override the User-Agent used for queries (default=None)')
    parser.add_argument('-n', '--norun', dest='norun', action='store_true',
                        help='Display summary of command arguments without '
                        'performing any actions')
    parser.add_argument('-v', '--verbose', dest='verbose',
                        action='store_true',
                        help='Print additional information to stderr')

    # Add cone arguments.
    parser.add_argument(
        '--start_index', type=int, metavar='start_index',
        help='Start with this cone in cone file'
        f' Default={conelist_defaults["start_index"]}')
    parser.add_argument(
        '--cone_limit', type=int, metavar='cone_limit',
        help='Maximum number of cones to query'
        f' Default={conelist_defaults["cone_limit"]}')

    return parser


def receiveSignal(signalNumber, frame):
    now = datetime.now()
    dtstr = now.strftime('%Y-%m-%d %H:%M:%S.%f')
    logging.warning(f'Received signal {signalNumber} at {dtstr}', file=sys.stderr,
                    flush=True)


def receiveSIGTERM(signalNumber, frame):
    now = datetime.now()
    dtstr = now.strftime('%Y-%m-%d %H:%M:%S.%f')
    logging.warning(f'Received signal {signalNumber} at {dtstr}', file=sys.stderr,
                    flush=True)
    faulthandler.dump_traceback()


def catch_signals():
    """
    Catch SIGHUP, SIGQUIT and SIGTERM to allow running in the background.

    When these signals are caught, a message will go to stderr.

    SIGTERM also cause a stack trace to be sent to stderr.
    """

    # register the signals to be caught
    if platform.system() != 'Windows':
        try:
            signal.signal(signal.SIGHUP, receiveSignal)
        except AttributeError as e:
            logging.warning(f'Warning: unable to add signal.SIGHUP handler: {repr(e)}')

        try:
            signal.signal(signal.SIGQUIT, receiveSignal)
        except AttributeError as e:
            logging.warning(f'Warning: unable to add signal.SIGQUIT handler: {repr(e)}')

    # SIGTERM should be available on Windows.
    try:
        signal.signal(signal.SIGTERM, receiveSignal)
    except AttributeError as e:
        logging.warning(f'Warning: unable to add signal.SIGTERM handler: {repr(e)}')


def apply_query_defaults(parsed_args, defaults):
    for k in defaults:
        if getattr(parsed_args, k) is None:
            setattr(parsed_args, k, defaults[k])


def _enable_requests_logging():
    import http.client as http_client

    http_client.HTTPConnection.debuglevel = 1
    requests_log = logging.getLogger("requests.packages.urllib3")
    requests_log.setLevel(logging.DEBUG)
    requests_log.propagate = True


def _init_logging(args):
    level = logging.WARNING

    if (args.verbose):
        level = logging.DEBUG
        _enable_requests_logging()

    logging.basicConfig(format='%(levelname)s: (%(asctime)s)    %(message)s',
                        level=level,
                        datefmt='%Y-%m-%d %H:%M:%S')



import re
from servicemon.query import compute_user_agent


def test_user_agent(capsys):
    custom_agent = 'MyUserAgent/4.3 (comment string)'
    ua = compute_user_agent(custom_agent)
    assert ua == custom_agent

    default_ua = compute_user_agent(None)
    expected = (
        r'servicemon\/[0-9]\.[0-9][^ ]* \(IVOA\-monitor https\:\/\/github.com\/NASA-NAVO\/servicemon\) Python/3\.')
    match = re.search(expected, default_ua)
    assert match



[
  {'base_name': 'great_archive_tap',
  'service_type': 'tap',
  'access_url': 'http://services.great_archive.net/tap_service',
  'adql':'''
  SELECT *
  FROM ivoa.obscore
  WHERE
    1=CONTAINS(POINT('ICRS', s_ra, s_dec),
              CIRCLE('ICRS', {}, {}, {} ))
  '''
  }
]



[
    {'base_name': 'HEASARC_xmmssc',
     'service_type': 'tap',
     'access_url': 'https://heasarc.gsfc.nasa.gov/xamin/vo/tap',
     'adql':'''
SELECT
   *
   FROM xmmssc
   WHERE 
      1=CONTAINS(POINT('ICRS', ra_unc, dec_unc),
                 CIRCLE('ICRS', {}, {}, {} ))
    '''
     }
]



from servicemon.plugin_support import AbstractResultWriter


class SampleResultWriter(AbstractResultWriter, plugin_name='sample-writer',
                         description='A valid writer that does nothing.'):

    def begin(self, **kwargs):
        pass

    def one_result(self, stats):
        pass

    def end(self):
        pass
